@@ -3,7 +3,7 @@
                            Coda File System
                               Release 5
 
-          Copyright (c) 1987-1999 Carnegie Mellon University
+          Copyright (c) 1987-2008 Carnegie Mellon University
                   Additional copyrights listed below
 
 This  code  is  distributed "AS IS" without warranty of any kind under
@@ -53,6 +53,7 @@ Pittsburgh, PA.
 #include <dllist.h>
 
 #include <rpc2/rpc2_addrinfo.h>
+#include <rpc2/secure.h>
 
 #ifndef HAVE_STRUCT_SOCKADDR_STORAGE
 /* this should be large enough to fit 'any' socket address. */
@@ -132,9 +133,11 @@ OBJ_HENTRY = 48127
 #define O_BINARY 0
 #endif
 
-/* MINRTO/MAXRTO are used to avoid unbounded timeouts */
-#define RPC2_MINRTO   300000        /* min rto (rtt + variance) is 300 msec */
-#define RPC2_MAXRTO   30000000     /* max rto (rtt + variance) is 30 seconds */
+/* MAXRTO is used to avoid unbounded timeouts */
+#define RPC2_MAXRTO     30000000   /* max rto (rtt + variance) is 30 seconds */
+#define RPC2_DELACK_DELAY 100000   /* delay for server to send an ack that it
+				      received a request. This provides an
+				      upper bound on server response time. */
 
 /* Definitions for Flags field of connections */
 #define CE_OLDV  0x1  /* old version detected during bind */
@@ -161,7 +164,7 @@ struct CEntry		/* describes a single RPC connection */
     /* State, identity  and sequencing */
     long State;
     RPC2_Handle  UniqueCID;
-    RPC2_Integer NextSeqNumber;	
+    RPC2_Unsigned NextSeqNumber;
     RPC2_Integer SubsysId;
     RPC2_Integer Flags;	    /* CE_OLDV ? */
     time_t LastRef;    			/* when CEntry was last looked up */
@@ -193,19 +196,17 @@ struct CEntry		/* describes a single RPC connection */
 					    for retransmission (reply or last packet
 					    of Bind handshake */
     unsigned long reqsize;              /* size of request, for RTT calcs */
-    unsigned long respsize;              /* size of response, for RTT calcs */
 
-    /* Retransmission info - client */
-#define LOWERLIMIT 300000               /* floor on lower limit, usec. */
-    unsigned long LowerLimit;           /* minimum retry interval, usec */
-
-    long RTT;                           /* Smoothed RTT estimate, 1msec units. */
-    long RTTVar;                        /* Variance of RTT, 1msec units. */
     unsigned int TimeStampEcho;
     unsigned int RequestTime;
-    long Retry_N;                       /* Number of retries for this connection. */
-    struct timeval *Retry_Beta;         /* Retry parameters for this connection. */
-    struct timeval SaveResponse;        /* 2*Beta0, lifetime of saved response packet. */
+    struct timeval TimeBomb;		/* Time limit for some rpc2 requests on
+					   this connection. */
+    struct timeval SaveResponse;        /* 2*KeepAlive, lifetime of saved response packet. */
+    RPC2_RequestFilter Filter;		/* Set on the server during binding,
+					   filter incoming requests so that the
+					   SubsysID/Connection matches that of
+					   the handler we authenticated with */
+    struct security_association sa;
     };
 
 
@@ -223,13 +224,6 @@ struct MEntry			/* describes an RPC multicast connection */
     struct RPC2_addrinfo    *ClientAddr;    /* |		*/
     RPC2_Handle		    MgroupID;	    /* |		*/
     RPC2_Integer	    NextSeqNumber;  /* for mgrp connection */
-    RPC2_Integer	    SubsysId;
-    RPC2_Integer	    Flags;
-
-    /* Security */
-    RPC2_Integer	    SecurityLevel;
-    RPC2_EncryptionKey	    SessionKey;
-    RPC2_Integer	    EncryptionType;
 
     /* Auxiliary stuff */
     struct SE_Definition *SEProcs;	/* pointer to side effect routines */
@@ -251,10 +245,6 @@ struct MEntry			/* describes an RPC multicast connection */
 #define	howmanylisteners    me_conns.me_client.mec_howmanylisteners
 #define	maxlisteners	    me_conns.me_client.mec_maxlisteners
 #define	conn		    me_conns.mes_conn
-
-    /* Other information - Only needed by client */
-    struct RPC2_addrinfo    *IPMAddr;	    /* IP Multicast Host Address */
-    RPC2_PacketBuffer	    *CurrentPacket; /* current multicast packet */
 	};
 
 
@@ -277,6 +267,9 @@ struct MEntry			/* describes an RPC multicast connection */
 		       destroyed by SocketListener on timeout
 	               destroyed by user LWP when reply is sent.
 
+          DELAYED_SEND Artificially introduced a transmission delay
+          DELAYED_RECV Artificially introduced a reception delay
+
           OTHER        associated with a specific connection
 	               created and destroyed by user LWP
 
@@ -287,12 +280,13 @@ connection they are associated with.
 */
 
 /* NOTE:  enum definitions  have to be non-anonymous: else a dbx bug is triggered */
-enum SL_Type {REPLY=1421, REQ=1422, OTHER=1423, DELACK=20010911};
+enum SL_Type {REPLY=1421, REQ=1422, OTHER=1423, DELACK=20010911,
+	      DELAYED_SEND=20061016, DELAYED_RECV=20061017};
 enum RetVal {WAITING=38358230, ARRIVED=38358231, TIMEOUT=38358232,
 	KEPTALIVE=38358233, KILLED=38358234, NAKED=38358235};
 
 /* data structure for communication with SocketListener */
-struct SL_Entry		
+struct SL_Entry
     {
     /* LinkEntry fields */
     struct SL_Entry *NextEntry;
@@ -300,19 +294,20 @@ struct SL_Entry
     enum {OBJ_SLENTRY = 107} MagicNumber;
     struct SL_Entry *Qname;
 
-    enum SL_Type Type;    
+    enum SL_Type Type;
 
     /* Timeout-related fields */
     struct TM_Elem TElem;	/* element  to be inserted into  timer chain;
-				    The BackPointer field of TElem will 
+				    The BackPointer field of TElem will
 				    point to this SL_Entry */
     enum RetVal ReturnCode;     /* SocketListener changes this from WAITING */
 
     /* Other fields */
     RPC2_Handle Conn;		/* NULL or conn corr to this SL Entry */
-    RPC2_PacketBuffer *Packet;  /* NULL,  awaiting retransmission or just arrived */
+    void *data;			/* NULL, packet buffer, or delayed xmit data */
     RPC2_RequestFilter Filter;  /* useful only in GetRequest */
-    long RetryIndex;		/* useful only in MakeRPC */
+    int RetryIndex;		/* useful only in MakeRPC */
+    struct timeval RInterval;	/* current retry timeout */
     };
 
 
@@ -327,7 +322,7 @@ struct SubsysEntry	/* Defines a subsystem being actively serviced by a server */
 
 
 /* extend with other side effect types */
-typedef enum {UNSET_HE = 0, RPC2_HE = 1, SMARTFTP_HE = 2} HEType;	
+typedef enum {UNSET_HE = 0, RPC2_HE = 1, SMARTFTP_HE = 2} HEType;
 
 struct HEntry {
     /* Link Entry Pointers */
@@ -347,15 +342,16 @@ struct HEntry {
     unsigned SE_NumEntries;	/* number of sideeffect observations recorded */
     RPC2_NetLogEntry SE_Log[RPC2_MAXLOGLENGTH];
 
+/* the RTT is shifted analogous to Jacobson's article in SIGCOMM'88 */
 #define RPC2_RTT_SHIFT    3     /* Bits to right of binary point of RTT */
-#define RPC2_RTTVAR_SHIFT 2     /* Bits to right of binary point of RTTVar */
-    unsigned long   RTT;	/* RTT          (us<<RPC2_RTT_SHIFT) */
-    unsigned long   RTTVar;	/* RTT variance (us<<RPC2_RTTVAR_SHIFT) */
+#define RPC2_RTTVAR_SHIFT 2     /* Bits to right of binary point of RTTvar */
+    unsigned long   RTT;	/* RTT		(us<<RPC2_RTT_SHIFT) */
+    unsigned long   RTTvar;	/* RTT variance	(us<<RPC2_RTTVAR_SHIFT) */
 
-#define RPC2_BR_SHIFT    3
-#define RPC2_BRVAR_SHIFT 2
-    unsigned long   BR;		/* Byterate          (ns/B<<RPC2_BW_SHIFT) */
-    unsigned long   BRVar;	/* Byterate variance (ns/B<<RPC2_BWVAR_SHIFT) */
+#define RPC2_INITIAL_BW 100000
+#define RPC2_BW_SHIFT    4
+    uint32_t BWlo_in,  BWhi_in;	/* Incoming Bandwidth	(B/s) */
+    uint32_t BWlo_out, BWhi_out; /* Outgoing Bandwidth	(B/s) */
 };
 
 
@@ -364,19 +360,33 @@ struct HEntry {
 /* WARNING: if you port RPC2, make sure the structure sizes work out to be the same on your target machine */
 
 struct Init1Body			/* Client to Server: format of packets with opcode of RPC2_INIT1xxx */
-    {
-    RPC2_NewConnectionBody FakeBody;	/* body of fake packet from RPC2_GetRequest() */
+{
+/* body of fake packet from RPC2_GetRequest() */
+    RPC2_Integer FakeBody_SideEffectType;
+    RPC2_Integer FakeBody_SecurityLevel;
+    RPC2_Integer FakeBody_EncryptionType;
+    RPC2_Integer FakeBody_AuthenticationType;
+    RPC2_Unsigned FakeBody_ClientIdent_SeqLen;
+    RPC2_Unsigned FakeBody_ClientIdent_SeqBody;
+/* end of body of fake packet from RPC2_GetRequest() */
+    /* When MakeFake is called we'll clobber the remaining data in this packet
+     * because we don't really send a pointer to the ClientIdent.SeqBody, but
+     * we move it there from the tail of the packet so that the stub generator
+     * can unpack it as a valid RPC2_NEWCONNECTION rpc call */
+
     RPC2_Integer XRandom;		/* encrypted random number */
     char usedtobehostport[92];		/* XXX not used anymore but old rpc2
 					   servers need the alignment */
     RPC2_Integer Uniquefier;		/* to allow detection of retransmissions */
-    RPC2_Integer Spare1;
-    RPC2_Integer Spare2;
+    RPC2_Unsigned RPC2SEC_version;	/* supported handshake version */
+    RPC2_Unsigned Preferred_Keysize;	/* preferred encryption key length */
     RPC2_Integer Spare3;
     RPC2_Byte Version[96];		/* set to RPC2_VERSION */
-    RPC2_Byte Text[4];	    /* Storage pointed to by FakeBody.ClientIdent.SeqBody;
-				4 bytes is a fake to cause proper longword alignment
-				on all interesting machines */
+    RPC2_Byte Text[4];			/* Storage for the ClientIdent. Moved
+					   to FakeBody_ClientIdent_SeqBody in
+					   rpc2a.c:MakeFake(). It can of course
+					   be more than 4 bytes, this is just
+					   for alignment purposes. */
     };
 
 struct Init2Body		/* Server to Client */
@@ -463,9 +473,8 @@ extern struct CBUF_Header *rpc2_TraceBuffHeader;
 extern PROCESS rpc2_SocketListenerPID;	/* used in IOMGR_Cancel() calls */
 extern unsigned long rpc2_LamportClock;
 extern long Retry_N;
-extern struct timeval *Retry_Beta;
-#define MaxRetryInterval Retry_Beta[0]
-extern struct timeval SaveResponse;
+extern struct timeval KeepAlive;
+extern uint32_t *rpc2_RTTvals;
 
 /* List manipulation routines */
 void rpc2_Replenish();
@@ -484,24 +493,23 @@ long rpc2_CreateIPSocket(int af, int *svar, struct RPC2_addrinfo *addr, short *P
 
 /* Packet  routines */
 long rpc2_SendReliably(), rpc2_MSendPacketsReliably();
-void rpc2_XmitPacket(RPC2_PacketBuffer *whichPB,
-		     struct RPC2_addrinfo *addr, int confirm);
+void rpc2_XmitPacket(RPC2_PacketBuffer *pb, struct RPC2_addrinfo *addr,
+		     int confirm);
 void rpc2_InitPacket();
 int rpc2_MorePackets(void);
 long rpc2_RecvPacket(long whichSocket, RPC2_PacketBuffer *whichBuff);
 void rpc2_htonp(RPC2_PacketBuffer *p);
 void rpc2_ntohp(RPC2_PacketBuffer *p);
-long rpc2_SetRetry(), rpc2_CancelRetry();
-void rpc2_ResetLowerLimit();
+long rpc2_CancelRetry(struct CEntry *Conn, struct SL_Entry *Sle);
 void rpc2_UpdateRTT(RPC2_PacketBuffer *pb, struct CEntry *ceaddr);
-void rpc2_ResetObs();
-void rpc2_ProcessPackets(), rpc2_ExpireEvents();
+void rpc2_ExpireEvents();
 
 /* Connection manipulation routines  */
 int rpc2_InitConn(void);
 void rpc2_FreeConn(), rpc2_SetConnError();
 struct CEntry *rpc2_AllocConn();
 struct CEntry *rpc2_ConnFromBindInfo(struct RPC2_addrinfo *peeraddr,
+				     RPC2_Handle RemoteHandle,
 				     RPC2_Integer whichUnique);
 struct CEntry *__rpc2_GetConn(RPC2_Handle handle); /* doesn't bump lastref */
 struct CEntry *rpc2_GetConn(RPC2_Handle handle);
@@ -520,9 +528,8 @@ void rpc2_ClearHostLog(struct HEntry *whichHost, NetLogEntryType type);
 
 void RPC2_UpdateEstimates(struct HEntry *whichHost, RPC2_Unsigned ElapsedTime,
 			  RPC2_Unsigned InBytes, RPC2_Unsigned OutBytes);
-void rpc2_RetryInterval(RPC2_Handle whichConn, RPC2_Unsigned InBytes,
-			RPC2_Unsigned OutBytes, int *retry, int maxretry,
-			struct timeval *tv);
+int rpc2_RetryInterval(struct CEntry *ce, int retry, struct timeval *tv,
+		       RPC2_Unsigned OutBytes, RPC2_Unsigned InBytes, int sftp);
 
 /* Multicast group manipulation routines */
 void rpc2_InitMgrp(), rpc2_FreeMgrp(), rpc2_RemoveFromMgrp(), rpc2_DeleteMgrp();
@@ -543,6 +550,7 @@ void rpc2_ClockTick(void *);
 /* rpc2_SocketListener packet handlers */
 void SL_RegisterHandler(unsigned int proto, void (*func)(RPC2_PacketBuffer *));
 void rpc2_HandlePacket(RPC2_PacketBuffer *pb);
+void RPC2_DispatchProcess(void);
 
 /* Packet timestamp creation */
 unsigned int rpc2_TVTOTS(const struct timeval *tv);
@@ -555,16 +563,18 @@ void rpc2_PrintTMElem(), rpc2_PrintFilter(), rpc2_PrintSLEntry(),
 	rpc2_PrintHostIdent(), rpc2_PrintPortIdent(), rpc2_PrintSEDesc();
 extern FILE *ErrorLogFile;
 
-unsigned int rpc2_TrueRandom(void);
-
 /* encryption */
+struct security_association *rpc2_GetSA(uint32_t spi);
+
 void rpc2_ApplyD(RPC2_PacketBuffer *pb, struct CEntry *ce);
 void rpc2_ApplyE(RPC2_PacketBuffer *pb, struct CEntry *ce);
 
 time_t rpc2_time();
 long rpc2_InitRetry(long HowManyRetries, struct timeval *Beta0);
 
-void rpc2_NoteBinding(struct RPC2_addrinfo *peeraddr, RPC2_Integer whichUnique, RPC2_Handle whichConn);
+void rpc2_NoteBinding(struct RPC2_addrinfo *peeraddr,
+		      RPC2_Handle RemoteHandle, RPC2_Integer whichUnique,
+		      RPC2_Handle whichConn);
 
 int mkcall(RPC2_HandleResult_func *ClientHandler, int ArgCount, int HowMany,
 	   RPC2_Handle ConnList[], long offset, long rpcval, int *args);
@@ -610,6 +620,8 @@ void rpc2_printaddrinfo(const struct RPC2_addrinfo *ai, FILE *f);
 void rpc2_splitaddrinfo(RPC2_HostIdent *Host, RPC2_PortIdent *Port,
 			const struct RPC2_addrinfo *ai);
 void rpc2_simplifyHost(RPC2_HostIdent *Host, RPC2_PortIdent *Port);
+void rpc2_formataddrinfo(const struct RPC2_addrinfo *ai, char *buf,
+			 size_t buflen, int use_canonname);
 
 
 /*--------------- Useful definitions that used to be in potpourri.h or util.h ---------------*/
@@ -683,4 +695,38 @@ do {\
 
 #define TSDELTA(_ts1_, _ts2_) ((int)(_ts1_) - (int)(_ts2_))
 
+#ifdef USE_LUA
+/* lua.c - include Lua interpreter for roundtrip time estimators */
+void LUA_init(void);
+void LUA_clocktick(void);
+void LUA_drop_hosttable(struct HEntry *he);
+
+void LUA_rtt_update(struct HEntry *he, uint32_t rtt, uint32_t tx, uint32_t rx);
+int LUA_rtt_getrto(struct HEntry *he, uint32_t tx, uint32_t rx);
+int LUA_rtt_getbandwidth(struct HEntry *he, uint32_t *bw_tx, uint32_t *bw_rx);
+/* not sure if we want to allow the estimator to control number of retries
+ * and/or overall timeout, so I might not actually use this function... */
+int LUA_rtt_retryinterval(struct HEntry *he, uint32_t n, uint32_t tx, uint32_t rx);
+int LUA_fail_delay(struct RPC2_addrinfo *addr, RPC2_PacketBuffer *pb, int out,
+		   struct timeval *tv);
+#else
+/* do not include Lua interpreter, define empty stubs */
+#define LUA_init()
+#define LUA_clocktick()
+#define LUA_drop_hosttable(a)
+#define LUA_rtt_update(a,b,c,d)
+#define LUA_rtt_getrto(a,b,c)		0
+#define LUA_rtt_getbandwidth(a,b,c)	0
+#define LUA_rtt_retryinterval(a,b,c,d)	0
+#define LUA_fail_delay(a,b,c,d)		0
 #endif
+
+int rpc2_DelayedSend(int s, struct RPC2_addrinfo *addr, RPC2_PacketBuffer *pb,
+		     struct timeval *delay);
+int rpc2_DelayedRecv(RPC2_PacketBuffer *pb, struct timeval *delay);
+
+void rpc2_SendDelayedPacket(struct SL_Entry *sl);
+RPC2_PacketBuffer *rpc2_RecvDelayedPacket(struct SL_Entry *sl);
+
+#endif /* _RPC2_PRIVATE_H_ */
+
